﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using GPSLib;

namespace GPSplug
{
    public partial class GPSplugForm : Form, GPSDataObserver
    {
        int chFix = -1;
        int chSpeed = -1;
        int chAltitude = -1;
        int chHeading = -1;

        bool lwChannelsCreated = false;
        public GPSSourceSerial com;
        GPSDataDispatcher dispatcher;
        private bool dispatcherStopping = false;

        private string UIValue_RMC_status = "";
        private string UIValue_RMC_time = "";
        private string UIValue_RMC_speed = "";
        private string UIValue_RMC_heading = "";
        private string UIValue_GGA_altitude = "";

        public GPSplugForm()
        {
            InitializeComponent();
            UpdatePortList();
            lwConnect();
            timerUpdatePortList.Enabled = true;
        }

        private void GPSplugForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (dispatcher != null)
            {
                dispatcher.stop();
                dispatcher = null;
            }
        }

        public void GPSSourceClosed()
        {
            com.Close();
            dispatcher = null;
            dispatcherStopping = true;
        }

        private void UpdatePortList()
        {
            string[] names = GPSSourceSerial.getPortNames();
            string selected = (String)listBoxInputPort.SelectedItem;
            listBoxInputPort.Items.Clear();
            foreach (string name in names)
            {
                listBoxInputPort.Items.Add(name);
            }
            if (selected != null)
            {
                listBoxInputPort.SelectedItem = selected;
            }
        }

        private void buttonUseInput_Click(object sender, EventArgs e)
        {
            if (com == null)
            {
                string inputSerialPortName = (String)listBoxInputPort.SelectedItem;
                if ((inputSerialPortName == null) || (inputSerialPortName.Length == 0))
                {
                    MessageBox.Show("Please decide on a serial port first");
                    return;
                }
                
                com = new GPSSourceSerial(inputSerialPortName);

                if (com.Open())
                {
                    timerUpdatePortList.Enabled = false;
                    buttonConnect.Text = "Disconnect";
                    dispatcher = new GPSDataDispatcher(com);
                    dispatcher.addObserver(this);
                    com.DiscardBuffer();
                    dispatcher.startup();
                    timerUpdateUI.Enabled = true;
                }
                else
                {
                    teardown();
                    MessageBox.Show("Could not connect to serial port " + inputSerialPortName);
                }
            }
            else
            {
                if (dispatcher != null)
                {
                    dispatcher.stop();
                    dispatcher = null;
                }
                else
                {
                    MessageBox.Show("Weird error: com is not null but dispatcher is!!");
                }
                teardown();
            }
        }

        private void teardown()
        {
            com.Close();
            com = null;
            if (dispatcher != null)
            {
                dispatcher.stop();
                dispatcher = null;
            }
            timerUpdateUI.Enabled = false;
            UpdatePortList();
            buttonConnect.Text = "Connect";
            timerUpdatePortList.Enabled = true;
            this.Text = "GPS plugin for LogWorks";
        }
        
        private void UpdateUIValues()
        {
            int t = com.Timeouts();
            if (t > 0)
            {
                this.Text = "GPS plugin for LogWorks   (timeouts: " + t.ToString() + ")";
            }
            else
            {
                this.Text = "GPS plugin for LogWorks";
            }
            textBoxStatus.Text = UIValue_RMC_status;
            textBoxTime.Text = UIValue_RMC_time;
            textBoxSpeed.Text = UIValue_RMC_speed;
            textBoxHeading.Text = UIValue_RMC_heading;
            textBoxAltitude.Text = UIValue_GGA_altitude;
            if (dispatcherStopping)
            {
                dispatcherStopping = false;
                teardown();
                MessageBox.Show("End of Data");
            }
        }

        private void SendToChannel(int channel, double value)
        {
            if (channel == -1)
            {
                return;
            }
            else
            {
                axLW2VDev1.SetValue(channel, value);
            }
        }

        public void GPSSourceUpdate(NMEAData data)
        {
            switch (data.type)
            {
                case NMEAData.types.RMC:
                    if (data.rmc.status == NMEADataRMC.StatusValues.Active)
                    {
                        SendToChannel(chFix, 1);
                    }
                    else
                    {
                        SendToChannel(chFix, 0);
                    }
                    SendToChannel(chSpeed, data.rmc.groundSpeedKMH());
                    SendToChannel(chHeading, data.rmc.heading);
                    UIValue_RMC_status = data.rmc.status.ToString();
                    UIValue_RMC_time = data.rmc.time.ToString();
                    UIValue_RMC_speed = data.rmc.groundSpeedKMH().ToString() + " km/h";
                    UIValue_RMC_heading = data.rmc.heading.ToString() + "°";
                    break;

                case NMEAData.types.GGA:
                    if (data.gga.isValid())
                    {
                        SendToChannel(chFix, 1);
                    }
                    else
                    {
                        SendToChannel(chFix, 0);
                    }
                    SendToChannel(chAltitude, data.gga.altitude);
                    UIValue_GGA_altitude = data.gga.altitude.ToString() + " m MSL";
                    break;

                default:
                    break;
            }
        }

        private void lwConnect()
        {
            if (!lwChannelsCreated)
            {
                if (axLW2VDev1.IsConnected == 0)
                {
                    string devname = "NMEA_GPS";    // this will show up in LogWorks
                    int devfunc = 0;                // we just increment it for each channel below
                    int devtype = 183;              // magic number, for caching channel setup

                    axLW2VDev1.AddChannelEx("Speed", "km/h", 0, 500, 0, devname, devtype, devfunc++, 0, ref chSpeed);
                    axLW2VDev1.AddChannelEx("Altitude", "m MSL", -500, 4500, 0, devname, devtype, devfunc++, 0, ref chAltitude);
                    axLW2VDev1.AddChannelEx("Heading", "°", 0, 360, 0, devname, devtype, devfunc++, 0, ref chHeading);
                    axLW2VDev1.AddChannelEx("Fix", "valid", 0, 1, 0, devname, devtype, devfunc++, 0, ref chFix);
                }
                else
                {
                    MessageBox.Show("Close Plugin and LogWorks and restart.\nStart Plugin before starting LogWorks.");
                    return;
                }

                if (chSpeed == -1 || chAltitude == -1 || chHeading == -1 || chFix == -1)
                {
                    MessageBox.Show("Failed creating LogWorks channels.\nClose Plugin and LogWorks and restart.\nStart Plugin before starting LogWorks.");
                    return;
                }
                else
                {
                    lwChannelsCreated = true;
                }
            }
        }

        private void buttonAbout_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Generic (NMEA) GPS plugin for Innovate LogWorks\n" +
                "(threaded version)\n\n" +
                "R. Andersson, A. Hedin 2009\n" +
                "\nFeel free to re-distribute, this is FREE software and you only get what you paid for.\n" +
                "\nhttp://www.abc.se/~rasmus/logworks/"
                );
        }

        private void timerUpdatePortList_Tick(object sender, EventArgs e)
        {
            UpdatePortList();
        }

        private void timerUpdateUI_Tick(object sender, EventArgs e)
        {
            UpdateUIValues();
        }
    }
}
